@@ -41,6 +41,8 @@ static int print_usage(void)
 	fprintf(stderr, "\t-u : print info of uuid tree only\n");
 	fprintf(stderr, "\t-b block_num : print info of the specified block"
                     " only\n");
+	fprintf(stderr,
+		"\t-t tree_id : print only the tree with the given id\n");
 	fprintf(stderr, "%s\n", BTRFS_BUILD_VERSION);
 	exit(1);
 }
@@ -136,12 +138,13 @@ int main(int ac, char **av)
 	int root_backups = 0;
 	u64 block_only = 0;
 	struct btrfs_root *tree_root_scan;
+	u64 tree_id = 0;
 
 	radix_tree_init();
 
 	while(1) {
 		int c;
-		c = getopt(ac, av, "deb:rRu");
+		c = getopt(ac, av, "deb:rRut:");
 		if (c < 0)
 			break;
 		switch(c) {
@@ -164,26 +167,37 @@ int main(int ac, char **av)
 			case 'b':
 				block_only = arg_strtou64(optarg);
 				break;
+			case 't':
+				tree_id = arg_strtou64(optarg);
+				break;
 			default:
 				print_usage();
 		}
 	}
+	set_argv0(av);
 	ac = ac - optind;
-	if (ac != 1)
+	if (check_argc_exact(ac, 1))
 		print_usage();
 
+	ret = check_arg_type(av[optind]);
+	if (ret != BTRFS_ARG_BLKDEV) {
+		fprintf(stderr, "'%s' is not a block device\n", av[optind]);
+		exit(1);
+	}
+
 	info = open_ctree_fs_info(av[optind], 0, 0, OPEN_CTREE_PARTIAL);
 	if (!info) {
 		fprintf(stderr, "unable to open %s\n", av[optind]);
 		exit(1);
 	}
+
 	root = info->fs_root;
+	if (!root) {
+		fprintf(stderr, "unable to open %s\n", av[optind]);
+		exit(1);
+	}
 
 	if (block_only) {
-		if (!root) {
-			fprintf(stderr, "unable to open %s\n", av[optind]);
-			exit(1);
-		}
 		leaf = read_tree_block(root,
 				      block_only,
 				      root->leafsize, 0);
@@ -207,7 +221,7 @@ int main(int ac, char **av)
 		goto close_root;
 	}
 
-	if (!(extent_only || uuid_tree_only)) {
+	if (!(extent_only || uuid_tree_only || tree_id)) {
 		if (roots_only) {
 			printf("root tree: %llu level %d\n",
 			     (unsigned long long)info->tree_root->node->start,
@@ -267,6 +281,8 @@ again:
 					      0);
 			if (!extent_buffer_uptodate(buf))
 				goto next;
+			if (tree_id && found_key.objectid != tree_id)
+				goto next;
 
 			switch(found_key.objectid) {
 			case BTRFS_ROOT_TREE_OBJECTID:
